/*
 * Copyright(C) 2016, Thibaud Briand, ENS Cachan <thibaud.briand@ens-cachan.fr>
 * Copyright(C) 2016, Jonathan Vacher, ENS Cachan <vacher@ceremade.dauphine.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
* @file filters.c
* @brief This file contains the functions that defines
* the proposed filters in the Fourier domain (see Section 4). These functions
* are used (in Algorithm 1) to compute the spectral samples of the
* proposed filters.
* In order to handle the parameters we use global variables which are declared
* below in this file (and are declared as extern variables in the header
* filters.h).
* @version 1.0
* @author Thibaud BRIAND <thibaud.briand@enpc.fr>;
* @author Jonathan VACHER <vacher@ceremade.dauphine.fr>;
*/

#include <math.h>
#include <complex.h>

/* M_PI is a POSIX definition */
#ifndef M_PI
/** macro definition for Pi */
#define M_PI 3.14159265358979323846
#endif
/* !M_PI */

/* parameters of filters (global variables) */
float global_shift_x;
float global_shift_y;
int global_derivative;
float global_gaussian;
int global_number_steer;
int global_steer;
float global_alpha;

/**
 * @brief Function in the Fourier domain that defines
 * the sinc filter (see Section 4.1).
 * @param x first coordinate position in the Fourier domain
 * @param y second coordinate position in the Fourier domain
 * @return A complex float value corresponding to the spectral
 * sample at point (x,y)
 */
complex float sinc_filter(float x, float y)
{
  (void)(x);
  (void)(y);
  return 1;
}

/**
 * @brief Function in the Fourier domain that defines
 * the shift filter (see Section 4.1.2). The global variables
 * global_shift_x and global_shift_y are the shifting parameters
 * (noted a1 and a2 in the article).
 * @param x first coordinate position in the Fourier domain
 * @param y second coordinate position in the Fourier domain
 * @return A complex float value corresponding to the spectral
 * sample at point (x,y)
 */
complex float shift_filter(float x, float y)
{
  float theta = global_shift_x*x+global_shift_y*y;
  return(cos(theta)+I*sin(theta));
}

/**
 * @brief Function in the Fourier domain that defines
 * the derivative filter (see Section 4.1.3). The global variable
 * global_derivative controls if the derivative is taken along
 * the x-axis (1) or the y-axis (0).
 * @param x first coordinate position in the Fourier domain
 * @param y second coordinate position in the Fourier domain
 * @return A complex float value corresponding to the spectral
 * sample at point (x,y)
 */
complex float derivative_filter(float x, float y)
{
  float tmp = global_derivative*x+(1-global_derivative)*y;
  return(I*tmp);
}

/**
 * @brief Function in the Fourier domain that defines
 * the laplacian filter (see Section 4.1.4).
 * @param x first coordinate position in the Fourier domain
 * @param y second coordinate position in the Fourier domain
 * @return A complex float value corresponding to the spectral
 * sample at point (x,y)
 */
complex float laplacian_filter(float x, float y)
{
  float tmp = -x*x-y*y;
  return tmp;
}

/**
 * @brief Function in the Fourier domain that defines
 * the derivative filter (see Section 4.1.5). The global variable
 * global_sigma corresponds to the standard deviation of the
 * gaussian filter (noted sigma in the article).
 * @param x first coordinate position in the Fourier domain
 * @param y second coordinate position in the Fourier domain
 * @return A complex float value corresponding to the spectral
 * sample at point (x,y)
 */
complex float gaussian_filter(float x, float y)
{
  /* polar definition */
  float r2 = x*x+y*y;
  float tmp = exp(-global_gaussian*global_gaussian*r2/2);
  return tmp;
}

/**
 * @brief Function in the Fourier domain that defines
 * the lowpass filter (see Section 4.1.6).
 * @param x first coordinate position in the Fourier domain
 * @param y second coordinate position in the Fourier domain
 * @return A complex float value corresponding to the spectral
 * sample at point (x,y)
 */
complex float lowpass_filter(float x, float y)
{
  /* polar definition */
  float r = hypot(x,y);
  if( r != 0 ) {
    float tmp = (r<=M_PI/4) +
            (r>M_PI/4)*(r<M_PI/2)*cos(M_PI/2*(log(4*r/M_PI)/log(2)));
    return tmp;
  }
  else{
    return 1;
  }
}

/**
 * @brief Function in the Fourier domain that defines
 * the highpass filter (see Section 4.1.6).
 * @param x first coordinate position in the Fourier domain
 * @param y second coordinate position in the Fourier domain
 * @return A complex float value corresponding to the spectral
 * sample at point (x,y)
 */
complex float highpass_filter(float x, float y)
{
  /* polar definition */
  float r = hypot(x,y);
  if( r > 0 ){
    float tmp = (r>=M_PI/2)+
            (r>M_PI/4)*(r<M_PI/2)*cos(M_PI/2*(log(2*r/M_PI)/log(2)));
    return tmp;
  }
  else{
    return 0;
  }
}

/**
 * @brief Function in the Fourier domain that defines
 * the lowpass filter (see Section 4.1.6). The global variables
 * global_numer_steer and global_steer correspond to the number
 * of orientation (noted Q in the article) and the orientation
 * of the filter (noted q in the article). The normalization constant
 * alpha_Q is pre-computed and stored in the global variable
 * global_alpha.
 * @param x first coordinate position in the Fourier domain
 * @param y second coordinate position in the Fourier domain
 * @return A complex float value corresponding to the spectral
 * sample at point (x,y)
 */
complex float steered_filter(float x, float y)
{
  int Q = global_number_steer;
  int q = global_steer;
  if(x==0 && y==0) return 1;
  else {
    /* polar definition */
    float theta,factor,cosinus_theta,sign;
    theta = atan2(x,y);
    factor = cos(theta-M_PI*q/Q);
    cosinus_theta = 1;
    sign = 1;

    /* computation of a power */
    for( int p = 1 ; p<Q ; p++) {
      cosinus_theta *= factor;
      sign *= -1;
    }

    float tmp = global_alpha*cosinus_theta*(
                (fabs(fmod(theta+3*M_PI-M_PI*q/Q,2*M_PI)-M_PI)<M_PI/2)+
                sign*(fabs(fmod(theta+2*M_PI-M_PI*q/Q,2*M_PI)-M_PI)<M_PI/2));
    return tmp;
  }
}

/**
 * @brief Compute the constant of reversibility alpha which is essential to
 * guarantee the reversibility of the steerable pyramid.
 * We use the product rule of log to compute this constant.
 * @param Q number of orientation bands
 * @return alpha_Q the constant of reversibility
 */
float compute_alpha(int Q)
{
    /* allocation */
    float log_fact_alpha1 = 0;
    float log_fact_alpha2 = 0;
    float log_alpha;
    float alpha_Q;

    /* compute the two sums */
    for(int k=2; k<Q; k++) log_fact_alpha1 += log(k);
    for(int k=2; k<2*Q-1; k++) log_fact_alpha2 += log(k);

    /* compute the log of alpha */
    log_alpha = (Q-1)*log(2)+ log_fact_alpha1
                - 0.5*(log(Q)+log_fact_alpha2);

    /* compute alpha */
    alpha_Q = exp(log_alpha);

    return(alpha_Q);
}
