/*
 * Copyright(C) 2016, Thibaud Briand, ENS Cachan <thibaud.briand@ens-cachan.fr>
 * Copyright(C) 2016, Jonathan Vacher, ENS Cachan <vacher@ceremade.dauphine.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file main_comparison.c
 * @brief Main function for the comparison (see Section 4.2.2) of the filtering
 * methods of Algorithm 1 for the nine filters proposed in Section 4.2.3.
 * Important modules are in the file @ref comparison.c. @n
 * 27 filtered images and the 27 difference images are computed
 * (3 per filter). @n
 * It also computes the modulus of the DFT input in logarithmic scale,
 * the boundary and total values and for every filter the difference values
 * (mean and max). The range (of the real part of
 * the filtered image with method 1) is computed for the 9 filters in order to
 * be able to compute the relative values and the relative differences.
 * The computed values are printed in the standard output. @n
 * The input image may be a color image. Algorithm 1 of the article is then
 * applied independently on each color channel. Consider an input image of
 * width w, height h and with pd number of channels. The DFT array and filtered
 * images corresponding to such an input are "complex-valued arrays" of size w,
 * h and pd. They are stored in an array of size w,h and 2*pd by spliting
 * the real part and the complex part. @n
 * The 27+27+1 = 55 saved images correspond to the real part of computed
 * images. An affine transformation is applied when the output format is not
 * "TIFF". @n
 * The reading and writing of images uses the iio library through the functions
 * iio_read_image_float_vec and iio_save_image_float_vec.
 * @version 1.0
 * @author Thibaud BRIAND <thibaud.briand@enpc.fr>;
 * @author Jonathan VACHER <vacher@ceremade.dauphine.fr>;
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "iio.h"
#include "comparison.h"
#include "toolbox.h"
#include "filters.h"

#define MAX_NB_CHAR 300
#define NB_FILTERED 27
#define NB_DIFF 27
#define NB_INFO 56
#define NB_RANGE 9

static int display_usage()
{
  printf("\nmain_comparison input -d directory/ -e extension \n");
  printf("-a a1 -b a2 -V variable -g sigma -Q Q -C C -D D -E E -M M\n");
  printf("\nRequired parameters :\n");
  printf("input : name of the input image\n");
  printf("\nOptionnal parameters :\n");
  printf("directory/ : output directory  (default current directory ./)\n");
  printf("extension : output extension type (default tiff)\n");
  printf("a1 : x-component shift (default 0.25)\n");
  printf("a2 : y-component shift (default 0.25)\n");
  printf("variable : choice of the partial derivative component ");
  printf("x <-> 1 and y <-> 0 (default 1)\n");
  printf("sigma : standard deviation of the Gaussian filter ");
  printf("(default sqrt(log(2))/pi)\n");
  printf("Q : number of orientations of the steered filters");
  printf("(default 4)\n");
  printf("C : multiplicative factor for visualization of difference ");
  printf("images (default 100000)\n");
  printf("D : multiplicative factor for visualization of DFT modulus ");
  printf("(default 15)\n");
  printf("E : multiplicative factor for visualization of high-pass ");
  printf("filtered images (default 5)\n");
  printf("M : new mean of filtered images for visualization when the ");
  printf("mean was 0 (default 127)\n");
  return(1);
}

/* Function for processing the options and parameters */
// c: pointer to original argc
// v: pointer to original argv
// o: option name after hyphen
// d: default value (if NULL, the option takes no argument)
static char *pick_option(int *c, char ***v, char *o, char *d)
{
  int argc = *c;
  char **argv = *v;
  int id = d ? 1 : 0;
  for (int i = 0; i < argc - id; i++)
    if (argv[i][0] == '-' && 0 == strcmp(argv[i] + 1, o)) {
      char *r = argv[i + id] + 1 - id;
      *c -= id + 1;
      for (int j = i; j < argc - id; j++)
        (*v)[j] = (*v)[j + id + 1];
      return r;
    }
  return d;
}


/* Main function for the comparison of the three filtering methods of Algorithm 1 for
 * 9 proposed filters (see Section 4.2.3). The computations are made in the function
 * comparison_all_filters. It first computes the boundary
 * value, the total value and the dft of the input image.
 * For each filter it computes :
 * - 3 filtered images
 * - 3 difference images
 * - 6 values (3 mean + 3 mean)
 * - 1 range
 * Finally it computes 27 filtered images and 27 difference images
 * indexed as follows :
 * 0-1-2    : sinc filter
 * 3-4-5    : shift filter
 * 6-7-8    : derivative filter
 * 9-10-11  : laplacian filter
 * 12-13-14 : gaussian filter
 * 15-16-17 : lowpass filter
 * 18-19-20 : highpass filter
 * 21-22-23 : horizontal filter
 * 24-25-26 : diagonal filter
 * The 56 computed values are stored with the following indices in print_info :
 * 0     : boundary value
 * 1     : total value
 * 2-7   : sinc values
 * 8-13  : shift values
 * 14-19 : derivative values
 * 20-25 : laplacian values
 * 26-31 : gaussian values
 * 32-37 : lowpass values
 * 38-43 : highpass values
 * 44-49 : horizontal values
 * 50-55 : diagonal values
 * The 9 range values are stored with the following indices :
 * 0 : sinc filter
 * 1 : shift filter
 * 2 : derivative filter
 * 3 : laplacian filter
 * 4 : gaussian filter
 * 5 : lowpass filter
 * 6 : highpass filter
 * 7 : horizontal filter
 * 8 : diagonal filter
 * The values are printed in the standard output. The real part of images are
 * saved using the iio library.
 */
int main(int argc, char *argv[])
{
  /* process the options and parameters */
  // output directory
  char *directory_name = pick_option(&argc, &argv, "d", "");
  // extension type
  char *extension = pick_option(&argc, &argv, "e", "tiff");
  // shift parameter x
  global_shift_x = atof(pick_option(&argc, &argv, "a", "0.25"));
  // shift parameter y
  global_shift_y = atof(pick_option(&argc, &argv, "b", "0.25"));
  // derivative parameter
  global_derivative = atoi(pick_option(&argc, &argv, "V", "1"));
  if( global_derivative != 0 && global_derivative != 1)
  {
    printf("Bad parameter for the derivative filter \n");
    return EXIT_FAILURE;
  }
  // std default value is sqrt(log(2))/pi (see Section 4.2.3)
  global_gaussian = atof(pick_option(&argc, &argv, "g", "0.26501036351"));
  if( global_gaussian <= 0)
  {
    printf("Bad parameter for the gaussian filter \n");
    return EXIT_FAILURE;
  }
  // number of orientations
  global_number_steer = atoi(pick_option(&argc, &argv, "Q", "4"));
  // output writting parameters
  float C = atof(pick_option(&argc, &argv, "C", "100000")); //difference factor
  float D = atof(pick_option(&argc, &argv, "D", "15")); // DFT factor
  float E = atof(pick_option(&argc, &argv, "E", "5")); // Highpass factor
  float M = atof(pick_option(&argc, &argv, "M", "127")); // Mean value

  /* display usage */
  if (argc != 2) {
        display_usage();
        return EXIT_FAILURE;
  }

  /* process the non-option parameter */
  char *filename_in = argv[1];

  /* read the input image */
  int w, h, pd;
  float *in = iio_read_image_float_vec(filename_in, &w, &h, &pd);

  /* declaration and memory allocation */
  float print_info[NB_INFO];
  float range_value[NB_RANGE];
  float *filtered_images[NB_FILTERED];
  for( int i=0; i<NB_FILTERED; i++)
    filtered_images[i] = malloc(2*w*h*pd*sizeof(float));
  float *diff_images[NB_DIFF];
  for( int i=0; i<NB_DIFF; i++)
    diff_images[i] = malloc(w*h*pd*sizeof(float));
  float *modulus_dft_in = malloc(w*h*pd*sizeof(float));
  float *out_image = malloc(w*h*pd*sizeof(float));
  char *fname_out = malloc(MAX_NB_CHAR*sizeof(char));

  /* computation of the images and the comparison values */
  comparison_all_filters(filtered_images, diff_images, modulus_dft_in,
                         print_info, range_value, in, w, h, pd);

  /* print the values to stdout */
    // index for the comparison values
    // +6 when considering the next filter
    int ind1;
    // index for the range value
    // +1 when considering the next filter
    int ind2;

  printf("BOUNDARY ENERGY : %f (%f%%)\n", print_info[0],
         print_info[0]/print_info[1]*100.0);

  printf("SINC FILTER\n");
  ind1 = 2;
  ind2 = 0;
  printf("DIFF MEAN 1 vs 2 : %f (%f%%) \n", print_info[ind1],
         print_info[ind1]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 2 : %f (%f%%) \n", print_info[ind1+1],
         print_info[ind1+1]/range_value[ind2]*100.0);
  printf("DIFF MEAN 1 vs 3 : %f (%f%%) \n", print_info[ind1+2],
         print_info[ind1+2]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 3 : %f (%f%%) \n", print_info[ind1+3],
         print_info[ind1+3]/range_value[ind2]*100.0);
  printf("DIFF MEAN 2 vs 3 : %f (%f%%) \n", print_info[ind1+4],
         print_info[ind1+4]/range_value[ind2]*100.0);
  printf("DIFF MAX 2 vs 3 : %f (%f%%) \n", print_info[ind1+5],
         print_info[ind1+5]/range_value[ind2]*100.0);

  printf("SHIFT FILTER\n");
  ind1 += 6;
  ind2 += 1;
  printf("DIFF MEAN 1 vs 2 : %f (%f%%) \n", print_info[ind1],
         print_info[ind1]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 2 : %f (%f%%) \n", print_info[ind1+1],
         print_info[ind1+1]/range_value[ind2]*100.0);
  printf("DIFF MEAN 1 vs 3 : %f (%f%%) \n", print_info[ind1+2],
         print_info[ind1+2]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 3 : %f (%f%%) \n", print_info[ind1+3],
         print_info[ind1+3]/range_value[ind2]*100.0);
  printf("DIFF MEAN 2 vs 3 : %f (%f%%) \n", print_info[ind1+4],
         print_info[ind1+4]/range_value[ind2]*100.0);
  printf("DIFF MAX 2 vs 3 : %f (%f%%) \n", print_info[ind1+5],
         print_info[ind1+5]/range_value[ind2]*100.0);

  printf("DERIVATIVE FILTER\n");
  ind1 += 6;
  ind2 += 1;
  printf("DIFF MEAN 1 vs 2 : %f (%f%%) \n", print_info[ind1],
         print_info[ind1]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 2 : %f (%f%%) \n", print_info[ind1+1],
         print_info[ind1+1]/range_value[ind2]*100.0);
  printf("DIFF MEAN 1 vs 3 : %f (%f%%) \n", print_info[ind1+2],
         print_info[ind1+2]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 3 : %f (%f%%) \n", print_info[ind1+3],
         print_info[ind1+3]/range_value[ind2]*100.0);
  printf("DIFF MEAN 2 vs 3 : %f (%f%%) \n", print_info[ind1+4],
         print_info[ind1+4]/range_value[ind2]*100.0);
  printf("DIFF MAX 2 vs 3 : %f (%f%%) \n", print_info[ind1+5],
         print_info[ind1+5]/range_value[ind2]*100.0);

  printf("LAPLACIAN FILTER\n");
  ind1 += 6;
  ind2 += 1;
  printf("DIFF MEAN 1 vs 2 : %f (%f%%) \n", print_info[ind1],
         print_info[ind1]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 2 : %f (%f%%) \n", print_info[ind1+1],
         print_info[ind1+1]/range_value[ind2]*100.0);
  printf("DIFF MEAN 1 vs 3 : %f (%f%%) \n", print_info[ind1+2],
         print_info[ind1+2]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 3 : %f (%f%%) \n", print_info[ind1+3],
         print_info[ind1+3]/range_value[ind2]*100.0);
  printf("DIFF MEAN 2 vs 3 : %f (%f%%) \n", print_info[ind1+4],
         print_info[ind1+4]/range_value[ind2]*100.0);
  printf("DIFF MAX 2 vs 3 : %f (%f%%) \n", print_info[ind1+5],
         print_info[ind1+5]/range_value[ind2]*100.0);

  printf("GAUSSIAN FILTER\n");
  ind1 += 6;
  ind2 += 1;
  printf("DIFF MEAN 1 vs 2 : %f (%f%%) \n", print_info[ind1],
         print_info[ind1]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 2 : %f (%f%%) \n", print_info[ind1+1],
         print_info[ind1+1]/range_value[ind2]*100.0);
  printf("DIFF MEAN 1 vs 3 : %f (%f%%) \n", print_info[ind1+2],
         print_info[ind1+2]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 3 : %f (%f%%) \n", print_info[ind1+3],
         print_info[ind1+3]/range_value[ind2]*100.0);
  printf("DIFF MEAN 2 vs 3 : %f (%f%%) \n", print_info[ind1+4],
         print_info[ind1+4]/range_value[ind2]*100.0);
  printf("DIFF MAX 2 vs 3 : %f (%f%%) \n", print_info[ind1+5],
         print_info[ind1+5]/range_value[ind2]*100.0);

  printf("LOWPASS FILTER\n");
  ind1 += 6;
  ind2 += 1;
  printf("DIFF MEAN 1 vs 2 : %f (%f%%) \n", print_info[ind1],
         print_info[ind1]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 2 : %f (%f%%) \n", print_info[ind1+1],
         print_info[ind1+1]/range_value[ind2]*100.0);
  printf("DIFF MEAN 1 vs 3 : %f (%f%%) \n", print_info[ind1+2],
         print_info[ind1+2]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 3 : %f (%f%%) \n", print_info[ind1+3],
         print_info[ind1+3]/range_value[ind2]*100.0);
  printf("DIFF MEAN 2 vs 3 : %f (%f%%) \n", print_info[ind1+4],
         print_info[ind1+4]/range_value[ind2]*100.0);
  printf("DIFF MAX 2 vs 3 : %f (%f%%) \n", print_info[ind1+5],
         print_info[ind1+5]/range_value[ind2]*100.0);

  printf("HIGHPASS FILTER\n");
  ind1 += 6;
  ind2 += 1;
  printf("DIFF MEAN 1 vs 2 : %f (%f%%) \n", print_info[ind1],
         print_info[ind1]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 2 : %f (%f%%) \n", print_info[ind1+1],
         print_info[ind1+1]/range_value[ind2]*100.0);
  printf("DIFF MEAN 1 vs 3 : %f (%f%%) \n", print_info[ind1+2],
         print_info[ind1+2]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 3 : %f (%f%%) \n", print_info[ind1+3],
         print_info[ind1+3]/range_value[ind2]*100.0);
  printf("DIFF MEAN 2 vs 3 : %f (%f%%) \n", print_info[ind1+4],
         print_info[ind1+4]/range_value[ind2]*100.0);
  printf("DIFF MAX 2 vs 3 : %f (%f%%) \n", print_info[ind1+5],
         print_info[ind1+5]/range_value[ind2]*100.0);

  printf("HORIZONTAL FILTER\n");
  ind1 += 6;
  ind2 += 1;
  printf("DIFF MEAN 1 vs 2 : %f (%f%%) \n", print_info[ind1],
         print_info[ind1]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 2 : %f (%f%%) \n", print_info[ind1+1],
         print_info[ind1+1]/range_value[ind2]*100.0);
  printf("DIFF MEAN 1 vs 3 : %f (%f%%) \n", print_info[ind1+2],
         print_info[ind1+2]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 3 : %f (%f%%) \n", print_info[ind1+3],
         print_info[ind1+3]/range_value[ind2]*100.0);
  printf("DIFF MEAN 2 vs 3 : %f (%f%%) \n", print_info[ind1+4],
         print_info[ind1+4]/range_value[ind2]*100.0);
  printf("DIFF MAX 2 vs 3 : %f (%f%%) \n", print_info[ind1+5],
         print_info[ind1+5]/range_value[ind2]*100.0);

  printf("DIAGONAL FILTER\n");
  ind1 += 6;
  ind2 += 1;
  printf("DIFF MEAN 1 vs 2 : %f (%f%%) \n", print_info[ind1],
         print_info[ind1]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 2 : %f (%f%%) \n", print_info[ind1+1],
         print_info[ind1+1]/range_value[ind2]*100.0);
  printf("DIFF MEAN 1 vs 3 : %f (%f%%) \n", print_info[ind1+2],
         print_info[ind1+2]/range_value[ind2]*100.0);
  printf("DIFF MAX 1 vs 3 : %f (%f%%) \n", print_info[ind1+3],
         print_info[ind1+3]/range_value[ind2]*100.0);
  printf("DIFF MEAN 2 vs 3 : %f (%f%%) \n", print_info[ind1+4],
         print_info[ind1+4]/range_value[ind2]*100.0);
  printf("DIFF MAX 2 vs 3 : %f (%f%%) \n", print_info[ind1+5],
         print_info[ind1+5]/range_value[ind2]*100.0);

  /* save the output files
   * when the output extension is not "tiff"
   * the writen image may be modified for
   * visualization purposes */
  /* Modulus DFT input */
  sprintf(fname_out,"%smod_dft.%s", directory_name, extension);
  if( 0 != strcmp(extension,"tiff"))
    affine_transformation(modulus_dft_in, 0, D, w*h*pd);
  iio_save_image_float_vec(fname_out, modulus_dft_in, w, h, pd);

  // index for the filtered and difference images
  // +3 when considering the next filter
  ind1 = 0;
  // index for the range value
  // +1 when considering the next filter
  ind2 = 0;

  /* sinc part */
    /* method 1 */
    sprintf(fname_out,"%ssinc1.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 2 */
    sprintf(fname_out,"%ssinc2.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+1], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 3 */
    sprintf(fname_out,"%ssinc3.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+2], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 1 vs 2 */
    sprintf(fname_out,"%ssinc_diff_12.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1], 0, C/range_value[ind2], w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1], w, h, pd);

    /* method 1 vs 3 */
    sprintf(fname_out,"%ssinc_diff_13.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1+1], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+1], w, h, pd);

    /* method 2 vs 3 */
    sprintf(fname_out,"%ssinc_diff_23.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1+2], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+2], w, h, pd);

  /* shift part */
  ind1 += 3;
  ind2 += 1;
    /* method 1 */
    sprintf(fname_out,"%sshift1.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 2 */
    sprintf(fname_out,"%sshift2.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+1], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 3 */
    sprintf(fname_out,"%sshift3.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+2], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 1 vs 2 */
    sprintf(fname_out,"%sshift_diff_12.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1], 0, C/range_value[ind2], w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1], w, h, pd);

    /* method 1 vs 3 */
    sprintf(fname_out,"%sshift_diff_13.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1+1], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+1], w, h, pd);

    /* method 2 vs 3 */
    sprintf(fname_out,"%sshift_diff_23.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1+2], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+2], w, h, pd);

  /* derivative part */
  ind1 += 3;
  ind2 += 1;
    /* method 1 */
    sprintf(fname_out,"%sderivative1.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1], w, h, pd);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(out_image, M, 1, w*h*pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 2 */
    sprintf(fname_out,"%sderivative2.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+1], w, h, pd);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(out_image, M, 1, w*h*pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 3 */
    sprintf(fname_out,"%sderivative3.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+2], w, h, pd);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(out_image, M, 1, w*h*pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 1 vs 2 */
    sprintf(fname_out,"%sderivative_diff_12.%s", directory_name, extension);
    if( 0 != strcmp(extension,"tiff"))
      affine_transformation(diff_images[ind1], 0, C/range_value[ind2], w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1], w, h, pd);

    /* method 1 vs 3 */
    sprintf(fname_out,"%sderivative_diff_13.%s", directory_name, extension);
    if( 0 != strcmp(extension,"tiff"))
      affine_transformation(diff_images[ind1+1], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+1], w, h, pd);

    /* method 2 vs 3 */
    sprintf(fname_out,"%sderivative_diff_23.%s", directory_name, extension);
    if( 0 != strcmp(extension,"tiff"))
      affine_transformation(diff_images[ind1+2], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+2], w, h, pd);

  /* laplacian part */
  ind1 += 3;
  ind2 += 1;
    /* method 1 */
    sprintf(fname_out,"%slaplacian1.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1], w, h, pd);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(out_image, M, 1, w*h*pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 2 */
    sprintf(fname_out,"%slaplacian2.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+1], w, h, pd);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(out_image, M, 1, w*h*pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 3 */
    sprintf(fname_out,"%slaplacian3.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+2], w, h, pd);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(out_image, M, 1, w*h*pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 1 vs 2 */
    sprintf(fname_out,"%slaplacian_diff_12.%s", directory_name, extension);
    if( 0 != strcmp(extension,"tiff"))
      affine_transformation(diff_images[ind1], 0, C/range_value[ind2], w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1], w, h, pd);

    /* method 1 vs 3 */
    sprintf(fname_out,"%slaplacian_diff_13.%s", directory_name, extension);
    if( 0 != strcmp(extension,"tiff"))
      affine_transformation(diff_images[ind1+1], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+1], w, h, pd);

    /* method 2 vs 3 */
    sprintf(fname_out,"%slaplacian_diff_23.%s", directory_name, extension);
    if( 0 != strcmp(extension,"tiff"))
      affine_transformation(diff_images[ind1+2], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+2], w, h, pd);

  /* gaussian part */
  ind1 += 3;
  ind2 += 1;
    /* method 1 */
    sprintf(fname_out,"%sgaussian1.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 2 */
    sprintf(fname_out,"%sgaussian2.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+1], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 3 */
    sprintf(fname_out,"%sgaussian3.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+2], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 1 vs 2 */
    sprintf(fname_out,"%sgaussian_diff_12.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1], 0, C/range_value[ind2], w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1], w, h, pd);

    /* method 1 vs 3 */
    sprintf(fname_out,"%sgaussian_diff_13.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1+1], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+1], w, h, pd);

    /* method 2 vs 3 */
    sprintf(fname_out,"%sgaussian_diff_23.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1+2], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+2], w, h, pd);

  /* lowpass part */
  ind1 += 3;
  ind2 += 1;
    /* method 1 */
    sprintf(fname_out,"%slowpass1.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 2 */
    sprintf(fname_out,"%slowpass2.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+1], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 3 */
    sprintf(fname_out,"%slowpass3.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+2], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 1 vs 2 */
    sprintf(fname_out,"%slowpass_diff_12.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1], 0, C/range_value[ind2], w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1], w, h, pd);

    /* method 1 vs 3 */
    sprintf(fname_out,"%slowpass_diff_13.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1+1], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+1], w, h, pd);

    /* method 2 vs 3 */
    sprintf(fname_out,"%slowpass_diff_23.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1+2], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+2], w, h, pd);

  /* highpass part */
  ind1 += 3;
  ind2 += 1;
    /* method 1 */
    sprintf(fname_out,"%shighpass1.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1], w, h, pd);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(out_image, M, E, w*h*pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 2 */
    sprintf(fname_out,"%shighpass2.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+1], w, h, pd);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(out_image, M, E, w*h*pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 3 */
    sprintf(fname_out,"%shighpass3.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+2], w, h, pd);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(out_image, M, E, w*h*pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 1 vs 2 */
    sprintf(fname_out,"%shighpass_diff_12.%s", directory_name, extension);
    if( 0 != strcmp(extension,"tiff"))
      affine_transformation(diff_images[ind1], 0, C/range_value[ind2], w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1], w, h, pd);

    /* method 1 vs 3 */
    sprintf(fname_out,"%shighpass_diff_13.%s", directory_name, extension);
    if( 0 != strcmp(extension,"tiff"))
      affine_transformation(diff_images[ind1+1], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+1], w, h, pd);

    /* method 2 vs 3 */
    sprintf(fname_out,"%shighpass_diff_23.%s", directory_name, extension);
    if( 0 != strcmp(extension,"tiff"))
      affine_transformation(diff_images[ind1+2], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+2], w, h, pd);

  /* horizontal part */
  ind1 += 3;
  ind2 += 1;
    /* method 1 */
    sprintf(fname_out,"%shorizontal1.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 2 */
    sprintf(fname_out,"%shorizontal2.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+1], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 3 */
    sprintf(fname_out,"%shorizontal3.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+2], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 1 vs 2 */
    sprintf(fname_out,"%shorizontal_diff_12.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1], 0, C/range_value[ind2], w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1], w, h, pd);

    /* method 1 vs 3 */
    sprintf(fname_out,"%shorizontal_diff_13.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1+1], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+1], w, h, pd);

    /* method 2 vs 3 */
    sprintf(fname_out,"%shorizontal_diff_23.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1+2], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+2], w, h, pd);

  /* diagonal part */
  ind1 += 3;
  ind2 += 1;
    /* method 1 */
    sprintf(fname_out,"%sdiagonal1.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 2 */
    sprintf(fname_out,"%sdiagonal2.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+1], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 3 */
    sprintf(fname_out,"%sdiagonal3.%s", directory_name, extension);
    real_part(out_image, filtered_images[ind1+2], w, h, pd);
    iio_save_image_float_vec(fname_out, out_image, w, h, pd);

    /* method 1 vs 2 */
    sprintf(fname_out,"%sdiagonal_diff_12.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1], 0, C/range_value[ind2], w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1], w, h, pd);

    /* method 1 vs 3 */
    sprintf(fname_out,"%sdiagonal_diff_13.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1+1], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+1], w, h, pd);

    /* method 2 vs 3 */
    sprintf(fname_out,"%sdiagonal_diff_23.%s", directory_name, extension);
    if( 0 != strcmp(extension, "tiff"))
      affine_transformation(diff_images[ind1+2], 0, C/range_value[ind2],
                            w*h*pd);
    iio_save_image_float_vec(fname_out, diff_images[ind1+2], w, h, pd);

  /* free memory */
  free(in);
  free(out_image);
  free(modulus_dft_in);
  free(fname_out);
  for( int i=0; i<NB_FILTERED; i++)
    free(filtered_images[i]);
  for( int i=0; i<NB_DIFF; i++)
    free(diff_images[i]);

  return EXIT_SUCCESS;
}
