/*
 * Copyright(C) 2016, Thibaud Briand, ENS Cachan <thibaud.briand@ens-cachan.fr>
 * Copyright(C) 2016, Jonathan Vacher, ENS Cachan <vacher@ceremade.dauphine.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file main_evaluation.c
 * @brief Main function for the filtering of images when the filter is defined
 * in the Fourier domain (Algorithm 1).
 * The implementation of the algorithm is in @ref filtering.c. @n
 * The input image may be a color image. Algorithm 1 of the article is then
 * applied independently on each color channel. Consider an input image of
 * width w, height h and with pd number of channels. The DFT array and filtered
 * images corresponding to such an input are "complex-valued arrays" of size w,
 * h and pd. They are stored in an array of size w,h and 2*pd by spliting
 * the real part and the complex part. @n
 * Images which are saved correspond to the real part of computed images.
 * The reading and writing of images uses the iio library through the functions
 * iio_read_image_float_vec and iio_save_image_float_vec.
 * @version 1.0
 * @author Thibaud BRIAND <thibaud.briand@enpc.fr>;
 * @author Jonathan VACHER <vacher@ceremade.dauphine.fr>;
 */

/**
* @mainpage Source code for the filtering algorithm.
* This sources code is a C99 implementation of the filtering
* algorithm proposed in the article
* "How to Apply a Filter Defined in the Frequency Domain"
* <a href="http://www.ipol.im/pub/art/2016/116/"
* title="Go to the IPOL webpage of the linear filtering">
* whose IPOL webpage accessible here</a>.
*
* This code is provided with an HTML documentation produced
* by doxygen <http://www.doxygen.org/>.
*
* @author Thibaud BRIAND <thibaud.briand@ens-cachan.fr>;
* @author Jonathan VACHER <vacher@ceremade.dauphine.fr>;
*
* @b Compilation: @n
*
* To compile this code cmake, make and a C99 compiler (ie. gcc) are required.
* The FFTW3, libpng and libtiff libraries must be installed in the system. @n
* The libjpeg library is optional.
*
* Compile the binaries by running the following line in the current directory:
*
* $ mkdir build; cd build; cmake ..; make
*
* the process will produce the binaries: main_evaluation and main_comparison.
* This documentation only deals with main_evaluation.
*
* @b Usage:
* Running the binaries without parameters will print usage instructions as: @n
* main_evaluation input output method filter [params] @n
* Required parameters @n
* input : name of the input image @n
* output : name of the output image @n
* method : method for the filtering algorithm (1,2 or 3) @n
* filter : name of the filter (see below for the possibilities)  @n
* Optionnal parameters : the filters have 0,1 or 2 parameters @n
* sinc : no parameter  @n
* shift : 2 parameters a1 a2 the shift values along both axis
*  (default 0.25 0.25) @n
* derivative : 1 parameter equals to 0 or 1 (default 1) @n
* 0 --> y-axis derivative @n
* 1 --> x-axis derivative @n
* laplacian : no parameter @n
* gaussian : 1 parameter for the standard deviation
* (default sqrt(log(2))/pi) @n
* lowpass : no parameter  @n
* highpass : no parameter  @n
* steered : 2 parameters Q and q  @n
* Q --> the number of orientation (default 4) @n
* q --> orientation of the filer 0<=q< Q (default 0) @n
*
* @b Output :
*
* The filtered image with the given method and filter.
*
* @b Test :
*
* To test the module run : @n
* build/main_evaluation data/lena_gray.png output.tiff 1 laplacian
*
* [vflip](<https://github.com/gfacciol/omniflip>) can be used to display
* TIFF images.
*
* @b Releases :
*
* @li Version 1.00 : original version.
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "iio.h"
#include "filtering.h"
#include "filters.h"

static int display_usage()
{
  printf("\nmain_evaluation input output method filter [params]\n");
  printf("Required parameters :\n");
  printf("input : name of the input image\n");
  printf("output : name of the output image\n");
  printf("method : method for the filtering algorithm (1,2 or 3)\n");
  printf("filter : name of the filter (see below for the possibilities) \n");
  printf("Optionnal parameters : the filters have 0,1 or 2 parameters\n");
  printf("sinc : no parameter \n");
  printf("shift : 2 parameters a1 a2 the shift values along both axis");
  printf(" (default 0.25 0.25)\n");
  printf("derivative : 1 parameter equals to 0 or 1 (default 1)\n");
  printf("0 --> y-axis derivative \n1 --> x-axis derivative\n");
  printf("laplacian : no parameter\n");
  printf("gaussian : 1 parameter for the standard deviation ");
  printf("(default sqrt(log(2))/pi)\n");
  printf("lowpass : no parameter \n");
  printf("highpass : no parameter \n");
  printf("steered : 2 parameters Q and q \n");
  printf("Q --> the number of orientation (default 4)\n");
  printf("q --> orientation of the filer 0<=q<Q (default 0)\n");
  return 1;
}

int main(int argc, char *argv[])
{

  /* display usage */
  if (argc < 5 || argc > 7) {
        display_usage();
        return EXIT_FAILURE;
  }

  /* process parameters */
  char *filename_in = argv[1];
  char *filename_out= argv[2];
  int method = atoi(argv[3]);
  char *filter = argv[4];

  /* filter selection */
  filter_t f;
  if( 0 == strcmp(filter,"sinc") )
  {
    f = sinc_filter;
  }
  else if( 0 == strcmp(filter,"shift") )
  {
    f = shift_filter;
    global_shift_x = (argc > 5) ? atof(argv[5]) : 0.25;
    global_shift_y = (argc > 6) ? atof(argv[6]) : 0.25;
  }
  else if( 0 == strcmp(filter,"derivative" ) )
  {
    f = derivative_filter;
    global_derivative = (argc > 5) ? atoi(argv[5]) : 1;
    if( global_derivative != 0 && global_derivative != 1)
    {
      printf("Bad parameter for the derivative filter \n");
      return EXIT_FAILURE;
    }
  }
  else if( 0 == strcmp(filter,"laplacian" ) )
  {
    f = laplacian_filter;
  }
  else if( 0 == strcmp(filter,"gaussian" ) )
  {
    f = gaussian_filter;
    global_gaussian = (argc > 5) ? atof(argv[5]) : 0.26501036351;
    // default value sqrt(log(2))/pi
    if( global_gaussian <= 0)
    {
      printf("Bad parameter for the gaussian filter \n");
      return EXIT_FAILURE;
    }
  }
  else if( 0 == strcmp(filter,"highpass" ) )
  {
    f = highpass_filter;
  }
  else if( 0 == strcmp(filter,"lowpass" ) )
  {
    f = lowpass_filter;
  }
  else if( 0 == strcmp(filter,"steered" ) )
  {
    f = steered_filter;
    global_number_steer = (argc > 5) ? atoi(argv[5]) : 4;
    global_steer = (argc > 6) ? atoi(argv[6]) : 0 ;
    global_alpha = compute_alpha(global_number_steer);
  }
  else
  {
    printf("Unknown filter type \n");
    return EXIT_FAILURE;
  }

  /* read the input image */
  int w, h, pd;
  float *in = iio_read_image_float_vec(filename_in, &w, &h, &pd);

  /* allocate memory for the filtered image */
  float *out = malloc(w*h*pd*sizeof(float));

  /* compute the real part of the filtered image */
  real_part_apply_filter(out, in, f, method, w, h, pd);

  /* Write output */
  iio_save_image_float_vec(filename_out, out, w, h, pd);

  /* free memory */
  free(in);
  free(out);

  return EXIT_SUCCESS;
}
