/*
 * Copyright(C) 2016, Thibaud Briand, ENS Cachan <thibaud.briand@enpc.fr>
 * Copyright(C) 2016, Jonathan Vacher, ENS Cachan <vacher@ceremade.dauphine.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
* @file toolbox.c
* @brief This file contains auxiliary functions for image transformations,
* range and FFT computations.
*
* @version 1.0
* @author Thibaud BRIAND <thibaud.briand@enpc.fr>;
* @author Jonathan VACHER <vacher@ceremade.dauphine.fr>;
*/

#include <stdlib.h>
#include <complex.h>
#include <fftw3.h>

#ifndef USE_WISDOM
void evoke_wisdom(void) {}
void bequeath_wisdom(void) {}
#else//USE_WISDOM
#include "fftwisdom.c"
#endif//USE_WISDOM

#define FORI(n) for(int i=0;i<(n);i++)
#define FORJ(n) for(int j=0;j<(n);j++)
#define FORL(n) for(int l=0;l<(n);l++)

/* Image transformation */

/**
 * @brief Affine transformation of an array.
 * This function is used to modify images when writing output to
 * an extension type that is not "tiff".
 * @param in input array (which is modified)
 * @param m shift parameter
 * @param C multiplicative parameter
 * @param N input length
 */
void affine_transformation(float *in, float m, float C, int N)
{
   FORI(N) in[i] = C*in[i]+m;
}

/**
 * @brief Select the real part of an array (interpreted as a complex one).
 * The filtered images are complex-valued but the outputs are only real
 * parts.
 * @param out input real part
 * @param in input array
 * @param w input width
 * @param h input height
 * @param pd input number of channels
 */
void real_part(float *out, float *in, int w, int h, int pd)
{
  FORL(pd){
        FORI(w*h) out[i*pd+l] = in[2*(i*pd + l)];
    }
}

/* Range */

/**
 * @brief Compute the minimum value of an array.
 *
 * @param in input array
 * @param N input length
 * @return the minimum value (min) of the array
 */
static float min(float *in, int N)
{
  float min = in[0];
  FORI(N)
     if ( in[i] < min ) min = in[i];
  return(min);
}

/**
 * @brief Compute the maximum value of an array.
 *
 * @param in input array
 * @param N input length
 * @return the maximum value (max) of the array
 */
float max(float *in, int N)
{
  float max = in[0];
  FORI(N)
     if ( in[i] > max ) max = in[i];
  return(max);
}

/**
 * @brief Compute the range of an array (max - min).
 * This function is used for computing relative differences values.
 * @param in input array
 * @param N input length
 * @return the range of the array
 */
float range(float *in, int N)
{
  float maxi = max(in,N);
  float mini = min(in,N);
  return(maxi - mini);
}

/* DFT and inverse DFT */

/**
 * @brief Wrapper around FFTW3 that computes the complex-valued DFT
 * of a real-valued image (gray-level).
 * @param fx complex-valued DFT
 * @param x real-valued image
 * @param w image width
 * @param h image height
 */
static void fft_2dfloat(fftwf_complex *fx, float *x, int w, int h)
{
    fftwf_complex *a = fftwf_malloc(w*h*sizeof*a);

    evoke_wisdom();
    fftwf_plan p = fftwf_plan_dft_2d(h, w, a, fx,
                                     FFTW_FORWARD, FFTW_ESTIMATE);
    bequeath_wisdom();

    FORI(w*h) a[i] = x[i]; // complex assignment!
    fftwf_execute(p);

    fftwf_destroy_plan(p);
    fftwf_free(a);
    fftwf_cleanup();
}

/**
 * @brief Wrapper around FFTW3 that computes the complex-valued iDFT
 * of a complex-valued frequantial image (gray-level).
 * @param ifx complex-valued iDFT
 * @param fx complex-valued image
 * @param w image width
 * @param h image height
 */
static void ifft_2dfloat(fftwf_complex *ifx,  fftwf_complex *fx, int w, int h)
{
    fftwf_complex *a = fftwf_malloc(w*h*sizeof*a);
    fftwf_complex *b = fftwf_malloc(w*h*sizeof*b);

    evoke_wisdom();
    fftwf_plan p = fftwf_plan_dft_2d(h, w, a, b,
                                     FFTW_BACKWARD, FFTW_ESTIMATE);
    bequeath_wisdom();

    FORI(w*h) a[i] = fx[i];
    fftwf_execute(p);

    /* rescaling */
    float scale = 1.0/((float) w*h);
    FORI(w*h) ifx[i] = b[i] * scale;

    fftwf_destroy_plan(p);
    fftwf_free(a);
    fftwf_free(b);
    fftwf_cleanup();
}

/**
 * @brief Compute the DFT of a color image.
 * @param y DFT array of the input
 * @param x input image
 * @param w input width
 * @param h input height
 * @param pd number of channels of the input
 */
void fft_direct(float *y, float *x, int w, int h, int pd)
{
    float *c = malloc(w*h*sizeof*c);
    fftwf_complex *gc = malloc(w*h*sizeof*gc);
    FORL(pd) {
        FORI(w*h)
            c[i] = x[i*pd + l];
        fft_2dfloat(gc, c, w, h);
        FORI(w*h) {
            y[2*(i*pd + l)+0] = creal(gc[i]);
            y[2*(i*pd + l)+1] = cimag(gc[i]);
        }
    }
    free(c);
    free(gc);
}

/**
 * @brief Compute the iDFT of a color DFT image.
 * @param y iDFT of the input (complex-valued !)
 * @param x the DFT input image
 * @param w input width
 * @param h input height
 * @param pd input number of channels
 */
void fft_inverse(float *y, float *x, int w, int h, int pd)
{
    int pdh = 2*pd;
    fftwf_complex *c = malloc(w*h*sizeof*c);
    fftwf_complex *gc = malloc(w*h*sizeof*gc);
    FORL(pd) {
        FORI(w*h)
            c[i] = x[i*pdh + 2*l] + I * x[i*pdh+2*l+1];
        ifft_2dfloat(gc, c, w, h);
        FORI(w*h){
            y[i*pdh + 2*l] = creal(gc[i]);
            y[i*pdh + 2*l+1] = cimag(gc[i]);
        }
    }
    free(c);
    free(gc);
}
